const coral = @import("coral");

const gfx = @import("../gfx.zig");

const std = @import("std");

buffers: [2]Buffer,
is_swapped: bool = false,
ref_count: AtomicCount = AtomicCount.init(1),
has_next: ?*Self = null,
has_prev: ?*Self = null,

const AtomicCount = std.atomic.Value(usize);

pub const Buffer = struct {
	arena: std.heap.ArenaAllocator,
	closed_handles: coral.stack.Sequential(usize),
	open_commands: coral.stack.Sequential(OpenCommand),
	draw_commands: coral.stack.Sequential(DrawCommand),
	close_commands: coral.stack.Sequential(CloseCommand),

	pub fn clear(self: *Buffer) void {
		self.close_commands.clear();
		self.draw_commands.clear();
		self.open_commands.clear();

		if (!self.arena.reset(.retain_capacity)) {
			std.log.warn("failed to reset the buffer of a gfx queue with retained capacity", .{});
		}
	}

	fn deinit(self: *Buffer) void {
		self.arena.deinit();
		self.closed_handles.deinit();
		self.open_commands.deinit();
		self.draw_commands.deinit();
		self.close_commands.deinit();
	}

	fn init(allocator: std.mem.Allocator) Buffer {
		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
			.closed_handles = .{.allocator = allocator},
			.open_commands = .{.allocator = allocator},
			.draw_commands = .{.allocator = allocator},
			.close_commands = .{.allocator = allocator},
		};
	}

	pub fn draw(self: *Buffer, command: DrawCommand) std.mem.Allocator.Error!void {
		try self.draw_commands.push(switch (command) {
			.instance_2d => |instance_2d| .{.instance_2d = instance_2d},
			.post_process => |post_process| .{.post_process = post_process},
		});
	}

	pub fn open(self: *Buffer, command: OpenCommand) std.mem.Allocator.Error!gfx.Handle {
		const reserved_handle = @as(gfx.Handle, switch (command.handle) {
			.none => @enumFromInt(reserve_handle: {
				if (self.closed_handles.get()) |handle| {
					std.debug.assert(self.closed_handles.pop());

					break: reserve_handle handle;
				}

				break: reserve_handle next_handle.fetchAdd(1, .monotonic);
			}),

			_ => |handle| handle,
		});

		std.debug.assert(reserved_handle != .none);

		const arena_allocator = self.arena.allocator();

		try self.open_commands.push(.{
			.resource = switch (command.resource) {
				.texture => |texture| .{
					.texture = .{
						.data = try arena_allocator.dupe(coral.io.Byte, texture.data),
						.width = texture.width,
						.format = texture.format,
						.access = texture.access,
					},
				},

				.mesh_2d => |mesh_2d| .{
					.mesh_2d = .{
						.indices = try arena_allocator.dupe(u16, mesh_2d.indices),
						.vertices = try arena_allocator.dupe(OpenCommand.Mesh2D.Vertex, mesh_2d.vertices),
					},
				},
			},

			.label = if (command.label) |label| try arena_allocator.dupe(coral.io.Byte, label) else null,
			.handle = reserved_handle,
		});

		return reserved_handle;
	}
};

pub const CloseCommand = struct {
	handle: gfx.Handle,
};

pub const DrawCommand = union (enum) {
	instance_2d: Instance,
	post_process: PostProcess,

	pub const Instance = struct {
		texture: gfx.Handle,
		mesh_2d: gfx.Handle,
		transform: gfx.Transform2D,
	};

	pub const PostProcess = struct {

	};
};

pub const OpenCommand = struct {
	handle: gfx.Handle = .none,
	label: ?[]const u8 = null,

	resource: union (enum) {
		texture: Texture,
		mesh_2d: Mesh2D,
	},

	pub const Mesh2D = struct {
		vertices: []const Vertex,
		indices: []const u16,

		pub const Vertex = struct {
			xy: gfx.Point2D,
			uv: gfx.Point2D,
		};
	};

	pub const Texture = struct {
		data: []const coral.io.Byte,
		width: u16,
		format: Format,
		access: Access,

		pub const Access = enum {
			static,
		};

		pub const Format = enum {
			rgba8888,
			bgra8888,
			argb8888,
			rgb888,
			bgr888,

			pub fn byte_size(self: Format) usize {
				return switch (self) {
					.rgba8888, .bgra8888, .argb8888 => 4,
					.rgb888, .bgr888 => 3,
				};
			}
		};
	};
};

pub fn acquire(self: *Self) void {
	self.ref_count.fetchAdd(1, .monotonic);
}

pub fn create() std.mem.Allocator.Error!*Self {
	const queue = try coral.heap.allocator.create(Self);

	errdefer coral.heap.allocator.destroy(queue);

	queue.* = .{
		.buffers = .{Buffer.init(coral.heap.allocator), Buffer.init(coral.heap.allocator)},
	};

	mutex.lock();

	defer mutex.unlock();

	if (has_tail) |tail| {
		tail.has_next = queue;
		queue.has_prev = tail;
	} else {
		std.debug.assert(has_head == null);

		has_head = queue;
	}

	has_tail = queue;

	return queue;
}

pub fn pending(self: *Self) *Buffer {
	return &self.buffers[@intFromBool(self.is_swapped)];
}

pub fn release(self: *Self) void {
	if (self.ref_count.fetchSub(1, .monotonic) == 1) {
		mutex.lock();

		defer mutex.unlock();

		if (self.has_prev) |prev| {
			prev.has_next = self.has_next;
		} else {
			has_head = self.has_next;
		}

		if (self.has_next) |next| {
			next.has_prev = self.has_prev;
		} else {
			has_tail = self.has_prev;
		}

		for (&self.buffers) |*buffer| {
			buffer.deinit();
		}

		coral.heap.allocator.destroy(self);
	}
}

pub fn submitted(self: *Self) *Buffer {
	return &self.buffers[@intFromBool(!self.is_swapped)];
}

const Self = @This();

var has_head = @as(?*Self, null);

var has_tail = @as(?*Self, null);

var mutex = std.Thread.Mutex{};

var next_handle = AtomicCount.init(1);

pub fn swap() void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;

	while (has_node) |node| : (has_node = node.has_next) {
		node.is_swapped = !node.is_swapped;

		node.pending().clear();
	}
}

pub fn visit_close_commands(visit: anytype, args: anytype) !void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;
	var iterations = @as(usize, 0);

	while (has_node) |node| : ({
		has_node = node.has_next;
		iterations += 1;
	}) {
		for (node.submitted().close_commands.values) |command| {
			try @call(.auto, visit, .{command} ++ args);
		}
	}
}

pub fn visit_draw_commands(visit: anytype, args: anytype) !void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;
	var iterations = @as(usize, 0);

	while (has_node) |node| : ({
		has_node = node.has_next;
		iterations += 1;
	}) {
		for (node.submitted().draw_commands.values) |command| {
			try @call(.auto, visit, .{command} ++ args);
		}
	}
}

pub fn visit_open_commands(visit: anytype, args: anytype) !void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;
	var iterations = @as(usize, 0);

	while (has_node) |node| : ({
		has_node = node.has_next;
		iterations += 1;
	}) {
		for (node.submitted().open_commands.values) |command| {
			try @call(.auto, visit, .{command} ++ args);
		}
	}
}
